# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Processing module to apply functions on merge (before and after), save and load."""

from cliconfig.processing.base import Processing

__all__ = ["Processing"]
